import unittest
from collections import UserDict

import pytest

from sickchill.oldbeard.numdict import NumDict


class NumDictTest(unittest.TestCase):
    """
    Test the NumDict class
    """

    def test_constructors(self):
        """
        Test NumDict constructors
        """
        # dicts for testing
        dict_0 = {}  # Empty dictionary
        dict_1 = {1: "Elephant"}  # Single numeric key
        dict_2 = {1: "Elephant", 2: "Mouse"}  # Multiple numeric keys
        dict_3 = {"3": "Aardvark"}  # Numeric string key
        dict_4 = {"3": "Aardvark", "4": "Ant"}  # Multiple numeric string keys
        dict_5 = {5: "Cat", "6": "Dog"}  # Mixed numeric and numeric string keys
        dict_6 = {1: None, "2": None}  # None as values
        dict_7 = {0: "Empty"}  # None as key

        # Construct NumDicts from dicts
        num_dict = NumDict()
        num_dict_0 = NumDict(dict_0)
        num_dict_1 = NumDict(dict_1)
        num_dict_2 = NumDict(dict_2)
        num_dict_3 = NumDict(dict_3)
        num_dict_4 = NumDict(dict_4)
        num_dict_5 = NumDict(dict_5)
        num_dict_6 = NumDict(dict_6)
        num_dict_7 = NumDict(dict_7)

        # Most NumDicts from dicts should compare equal...
        assert num_dict == {}
        assert num_dict_0 == dict_0
        assert num_dict_1 == dict_1
        assert num_dict_2 == dict_2

        # ...however, numeric keys are not equal to numeric string keys...
        assert num_dict_3 != dict_3
        assert num_dict_4 != dict_4
        assert num_dict_5 != dict_5
        assert num_dict_6 != dict_6

        assert num_dict_7 == dict_7

        # Construct dicts from NumDicts
        dict_from_num_dict = dict(num_dict)
        dict_from_num_dict_1 = dict(num_dict_1)
        dict_from_num_dict_2 = dict(num_dict_2)
        dict_from_num_dict_3 = dict(num_dict_3)
        dict_from_num_dict_4 = dict(num_dict_4)
        dict_from_num_dict_5 = dict(num_dict_5)
        dict_from_num_dict_6 = dict(num_dict_6)
        dict_from_num_dict_7 = dict(num_dict_7)

        # All dicts from NumDicts should compare equal
        assert num_dict == dict_from_num_dict
        assert num_dict_1 == dict_from_num_dict_1
        assert num_dict_2 == dict_from_num_dict_2
        assert num_dict_3 == dict_from_num_dict_3
        assert num_dict_4 == dict_from_num_dict_4
        assert num_dict_5 == dict_from_num_dict_5
        assert num_dict_6 == dict_from_num_dict_6
        assert num_dict_7 == dict_from_num_dict_7

        # Construct NumDicts from NumDicts
        num_dict_from_num_dict = NumDict(num_dict)
        num_dict_from_num_dict_0 = NumDict(num_dict_0)
        num_dict_from_num_dict_1 = NumDict(num_dict_1)
        num_dict_from_num_dict_2 = NumDict(num_dict_2)
        num_dict_from_num_dict_3 = NumDict(num_dict_3)
        num_dict_from_num_dict_4 = NumDict(num_dict_4)
        num_dict_from_num_dict_5 = NumDict(num_dict_5)
        num_dict_from_num_dict_6 = NumDict(num_dict_6)
        num_dict_from_num_dict_7 = NumDict(num_dict_7)

        # All NumDicts from NumDicts should compare equal
        assert num_dict == num_dict_from_num_dict
        assert num_dict_0 == num_dict_from_num_dict_0
        assert num_dict_1 == num_dict_from_num_dict_1
        assert num_dict_2 == num_dict_from_num_dict_2
        assert num_dict_3 == num_dict_from_num_dict_3
        assert num_dict_4 == num_dict_from_num_dict_4
        assert num_dict_5 == num_dict_from_num_dict_5
        assert num_dict_6 == num_dict_from_num_dict_6
        assert num_dict_7 == num_dict_from_num_dict_7

        # keyword arg constructor should fail
        with self.assertRaises(TypeError):
            NumDict(one=1, two=2)  # Raise TypeError since we can't have numeric keywords

        # item sequence constructors work fine...
        assert NumDict([(1, "Elephant"), (2, "Mouse")]) == dict_from_num_dict_2
        assert NumDict(dict=[(1, "Elephant"), (2, "Mouse")]) == dict_from_num_dict_2
        assert NumDict([(1, "Elephant"), ("2", "Mouse")]) == dict_from_num_dict_2
        assert NumDict(dict=[("1", "Elephant"), (2, "Mouse")]) == dict_from_num_dict_2

        # ...unless you have a non-numeric key
        with self.assertRaises(TypeError):
            NumDict([("Rat", 11), ("Snake", 12)])
        with self.assertRaises(TypeError):
            NumDict(dict=[("Rat", 11), ("Snake", 12)])

        # combining item sequence constructors with keyword args does not work
        with self.assertRaises(TypeError):  # Raise TypeError since we can't have numeric keywords
            NumDict([(1, "one"), (2, "two")], two=3, five=4)

        # alternate constructors
        dict_8 = {1: "Echo", 2: "Echo"}

        assert NumDict.fromkeys("1 2".split()) == dict_from_num_dict_6
        assert NumDict().fromkeys("1 2".split()) == dict_from_num_dict_6
        assert NumDict.fromkeys("1 2".split(), "Echo") == dict_8
        assert NumDict().fromkeys("1 2".split(), "Echo") == dict_8
        assert num_dict_1.fromkeys("1 2".split()) is not num_dict_1
        self.assertIsInstance(num_dict_1.fromkeys("1 2".split()), NumDict)
        self.assertIsInstance(num_dict_2.fromkeys("1 2".split()), NumDict)
        self.assertIsInstance(num_dict_3.fromkeys("1 2".split()), NumDict)
        self.assertIsInstance(num_dict_4.fromkeys("1 2".split()), NumDict)

    def test_repr(self):
        """
        Test representation of NumDicts
        """
        # dicts for testing
        dict_0 = {}  # Empty dictionary
        dict_1 = {1: "Elephant"}  # Single numeric key
        dict_2 = {1: "Elephant", 2: "Mouse"}  # Multiple numeric keys
        dict_3 = {"3": "Aardvark"}  # Numeric string key
        dict_4 = {"3": "Aardvark", "4": "Ant"}  # Multiple numeric string keys
        dict_5 = {5: "Cat", "6": "Dog"}  # Mixed numeric and numeric string keys
        dict_6 = {1: None, "2": None}  # None as values
        dict_7 = {0: "Empty"}  # None as key

        #  Construct NumDicts from dicts
        num_dict = NumDict()
        num_dict_0 = NumDict(dict_0)
        num_dict_1 = NumDict(dict_1)
        num_dict_2 = NumDict(dict_2)
        num_dict_3 = NumDict(dict_3)
        num_dict_4 = NumDict(dict_4)
        num_dict_5 = NumDict(dict_5)
        num_dict_6 = NumDict(dict_6)
        num_dict_7 = NumDict(dict_7)

        reps = (
            "{}",
            "{1: 'Elephant'}",
            "{1: 'Elephant', 2: 'Mouse'}",
            "'3': 'Aardvark'",
            "{'3': 'Aardvark', '4': 'Ant'}",
            "{5: 'Cat', '6': 'Dog'}",
            "{1: None, '2': None}",
            "{0: 'Empty'}",
        )

        # Most representations of NumDicts should compare equal to dicts...
        assert str(num_dict) == str({})
        assert repr(num_dict) == repr({})
        assert repr(num_dict) in reps

        assert str(num_dict_0) == str(dict_0)
        assert repr(num_dict_0) == repr(dict_0)
        assert repr(num_dict_0) in reps

        assert str(num_dict_1) == str(dict_1)
        assert repr(num_dict_1) == repr(dict_1)
        assert repr(num_dict_1) in reps

        assert str(num_dict_2) == str(dict_2)
        assert repr(num_dict_2) == repr(dict_2)
        assert repr(num_dict_2) in reps

        # ...however, numeric keys are not equal to numeric string keys...
        # ...so the string representations for those are different...
        assert str(num_dict_3) != str(dict_3)
        assert repr(num_dict_3) != repr(dict_3)
        assert repr(num_dict_3) not in reps

        assert str(num_dict_4) != str(dict_4)
        assert repr(num_dict_4) != repr(dict_4)
        assert repr(num_dict_4) not in reps

        assert str(num_dict_5) != str(dict_5)
        assert repr(num_dict_5) != repr(dict_5)
        assert repr(num_dict_5) not in reps

        assert str(num_dict_6) != str(dict_6)
        assert repr(num_dict_6) != repr(dict_6)
        assert repr(num_dict_6) not in reps

        # ...but None keys work just fine
        assert str(num_dict_7) == str(dict_7)
        assert repr(num_dict_7) == repr(dict_7)
        assert repr(num_dict_7) in reps

    def test_rich_comparison_and_len(self):
        """
        Test rich comparison and length
        """
        # dicts for testing
        dict_0 = {}  # Empty dictionary
        dict_1 = {1: "Elephant"}  # Single numeric key
        dict_2 = {1: "Elephant", 2: "Mouse"}  # Multiple numeric keys

        # Construct NumDicts from dicts
        num_dict = NumDict()
        num_dict_0 = NumDict(dict_0)
        num_dict_1 = NumDict(dict_1)
        num_dict_2 = NumDict(dict_2)

        # Construct NumDicts from NumDicts
        num_dict_from_num_dict = NumDict(num_dict)
        num_dict_from_num_dict_0 = NumDict(num_dict_0)
        num_dict_from_num_dict_1 = NumDict(num_dict_1)
        num_dict_from_num_dict_2 = NumDict(num_dict_2)

        all_dicts = [
            dict_0,
            dict_1,
            dict_2,
            num_dict,
            num_dict_0,
            num_dict_1,
            num_dict_2,
            num_dict_from_num_dict,
            num_dict_from_num_dict_0,
            num_dict_from_num_dict_1,
            num_dict_from_num_dict_2,
        ]
        for val_a in all_dicts:
            for val_b in all_dicts:
                assert (val_a == val_b) == (len(val_a) == len(val_b))

    def test_dict_access_and_mod(self):
        """
        Test num dict access and modification
        """
        # dicts for testing
        dict_0 = {}  # Empty dictionary
        dict_1 = {1: "Elephant"}  # Single numeric key
        dict_2 = {1: "Elephant", 2: "Mouse"}  # Multiple numeric keys

        #  Construct NumDicts from dicts
        num_dict_0 = NumDict()
        num_dict_1 = NumDict(dict_1)
        num_dict_2 = NumDict(dict_2)

        # test __getitem__
        assert num_dict_2[1] == "Elephant"
        with pytest.raises(KeyError):
            _ = num_dict_1["Mouse"]  # key is not numeric
        with pytest.raises(KeyError):
            _ = num_dict_1.__getitem__("Mouse")  # key is not numeric
        with pytest.raises(KeyError):
            _ = num_dict_1[None]  # key does not exist
        with pytest.raises(KeyError):
            _ = num_dict_1.__getitem__(None)  # key does not exist

        # Test __setitem__
        num_dict_3 = NumDict(num_dict_2)
        assert num_dict_2 == num_dict_3

        num_dict_3[2] = "Frog"
        assert num_dict_2 != num_dict_3

        # Check None keys and numeric key conversion
        num_dict_3["3"] = "Armadillo"
        num_dict_3[None] = "Cockroach"

        # Check long ints
        num_dict_3[12390809518259081208909880312] = "Squid"
        num_dict_3["12390809518259081208909880312"] = "Octopus"
        assert num_dict_3[12390809518259081208909880312] == "Octopus"

        with pytest.raises(TypeError):
            num_dict_3.__setitem__("Gorilla", 1)  # key is not numeric
        with pytest.raises(TypeError):
            num_dict_3["Chimpanzee"] = 1  # key is not numeric
        with pytest.raises(TypeError):
            num_dict_3[(4, 1)] = 1  # key is not numeric
        with pytest.raises(TypeError):
            num_dict_3[[1, 3, 4]] = 1  # key is not numeric and is not hashable

        # Test __delitem__
        del num_dict_3[3]
        del num_dict_3[None]
        with pytest.raises(KeyError):
            del num_dict_3[3]  # already deleted
        with pytest.raises(KeyError):
            num_dict_3.__delitem__(3)  # already deleted
        with pytest.raises(KeyError):
            del num_dict_3["Mouse"]  # key would not exist, since it is not numeric

        # Test clear
        num_dict_3.clear()
        assert num_dict_3 == {}

        # Test copy()
        num_dict_2a = dict_2.copy()
        assert num_dict_2 == num_dict_2a
        num_dict_2b = num_dict_2.copy()
        assert num_dict_2b == num_dict_2
        num_dict_2c = UserDict({1: "Elephant", 2: "Mouse"})
        num_dict_2d = num_dict_2c.copy()  # making a copy of a UserDict is special cased
        assert num_dict_2c == num_dict_2d

        class MyNumDict(NumDict):
            """
            subclass Numdict for testing
            """

            def display(self):
                """
                add a method to subclass to differentiate from superclass
                """
                print("MyNumDict:", self)

        my_num_dict = MyNumDict(num_dict_2)
        my_num_dict_a = my_num_dict.copy()
        assert my_num_dict_a == my_num_dict

        my_num_dict[1] = "Frog"
        assert my_num_dict_a != my_num_dict

        # Test keys, items, values
        assert sorted(num_dict_2) == sorted(dict_2)
        assert sorted(num_dict_2.items()) == sorted(dict_2.items())
        assert sorted(num_dict_2.values()) == sorted(dict_2.values())

        # Test "in".
        for i in num_dict_2:
            assert i in num_dict_2
            assert (i in num_dict_1) == (i in dict_1)
            assert (i in num_dict_0) == (i in dict_0)

        assert None not in num_dict_2
        assert (None in num_dict_2) is (None in dict_2)

        dict_2[None] = "Cow"
        num_dict_2[None] = dict_2[None]
        assert None in num_dict_2
        assert (None in num_dict_2) is (None in dict_2)

        assert "Penguin" not in num_dict_2

        # Test update
        test = NumDict()
        test.update(dict_2)
        assert test == num_dict_2

        # Test get
        for i in num_dict_2:
            assert num_dict_2.get(i) == num_dict_2[i]
            assert num_dict_1.get(i) == dict_1.get(i)
            assert num_dict_0.get(i) == dict_0.get(i)

        for i in ["purple", None, 12312301924091284, 23]:
            assert num_dict_2.get(i) == dict_2.get(i), i

        with pytest.raises(AssertionError):
            assert num_dict_2.get("1") == dict_2.get("1"), 1  # dict_2 expects string key which does not exist

        # Test "in" iteration.
        num_dict_2b = num_dict_2
        for i in range(20):
            num_dict_2[i] = str(i)
            num_dict_2b[str(i)] = str(i)
        assert num_dict_2 == num_dict_2b

        ikeys = []
        for k in num_dict_2:
            ikeys.append(k)
        assert set(ikeys) == set(num_dict_2)

        # Test setdefault
        val = 1
        test = NumDict()
        assert test.setdefault(val, 42) == 42
        assert test.setdefault(val, "42") == 42
        assert test.setdefault(val, 42) != "42"
        assert test.setdefault(val, "42") != "42"
        assert val in test

        assert test.setdefault(val, 23) == 42
        assert test.setdefault(val, "23") ==  42
        assert test.setdefault(val, 23) != "42"
        assert test.setdefault(val, "23") != "42"
        assert val in test

        # Test pop
        val = 1
        test = NumDict({val: 42})
        assert test.pop(val) == 42
        pytest.raises(KeyError, test.pop, val)
        assert test.pop(val, 1) == 1
        test[val] = 42
        assert test.pop(val, 1) == 42

        # Test popitem
        val = 1
        test = NumDict({val: 42})
        assert test.popitem() == (val, 42)
        pytest.raises(KeyError, test.popitem)

    def test_missing(self):
        """
        Test missing keys
        """
        # Make sure NumDict doesn't have a __missing__ method
        assert not hasattr(NumDict, "__missing__")

        class NumDictD(NumDict):
            """
            subclass defines __missing__ method returning a value
            """

            def __missing__(self, key):
                key = 42
                return key

        num_dict_d = NumDictD({1: 2, 3: 4})
        assert num_dict_d[1] == 2
        assert num_dict_d[3] == 4
        assert 2 not in num_dict_d
        assert 2 not in num_dict_d
        assert num_dict_d[2] == 42

        class NumDictE(NumDict):
            """
            subclass defines __missing__ method raising RuntimeError
            """

            def __missing__(self, key):
                raise RuntimeError(key)

        num_dict_e = NumDictE()
        try:
            num_dict_e[42]
        except RuntimeError as err:
            assert err.args == (42,)
        else:
            self.fail("num_dict_e[42] didn't raise RuntimeError")

        class NumDictF(NumDict):
            """
            subclass sets __missing__ instance variable (no effect)
            """

            def __init__(self):
                # An instance variable __missing__ should have no effect
                self.__missing__ = lambda key: None
                NumDict.__init__(self)

        num_dict_f = NumDictF()
        try:
            num_dict_f[42]
        except KeyError as err:
            assert err.args == (42,)
        else:
            self.fail("num_dict_f[42] didn't raise KeyError")

        class NumDictG(NumDict):
            """
            subclass doesn't define __missing__ at a all
            """

            pass

        num_dict_g = NumDictG()
        try:
            num_dict_g[42]
        except KeyError as err:
            assert err.args == (42,)
        else:
            self.fail("num_dict_g[42] didn't raise KeyError")

        class NumDictH(NumDictD):
            """
            subclass calls super classes __missing__ and modifies the value before returning it
            """

            def __missing__(self, key):
                return super().__missing__(key) + 1

        num_dict_h = NumDictH()
        assert num_dict_h[None] == num_dict_d[None] + 1


def test_main():
    """
    Run tests when run as main
    """
    import logging

    log = logging.getLogger(__name__)
    logging.basicConfig(level=logging.DEBUG)

    log.info("=======================")
    log.info("STARTING - COMMON TESTS")
    log.info("=======================")
    log.info("######################################################################")

    suite = unittest.TestLoader().loadTestsFromTestCase(NumDictTest)
    unittest.TextTestRunner(verbosity=2).run(suite)


if __name__ == "__main__":
    test_main()



import json
import os

from tornado.escape import xhtml_unescape

from sickchill.oldbeard.browser import foldersAtPath

from .index import WebRoot
from .routes import Route


@Route("/browser(/?.*)", name="filebrowser")
class WebFileBrowser(WebRoot):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def index(self, path="", includeFiles=False, fileTypes=""):

        self.set_header("Cache-Control", "max-age=0,no-cache,no-store")
        self.set_header("Content-Type", "application/json")

        return json.dumps(foldersAtPath(xhtml_unescape(path), True, bool(int(includeFiles)), fileTypes.split(",")))

    def complete(self, term, includeFiles=False, fileTypes=""):

        self.set_header("Cache-Control", "max-age=0,no-cache,no-store")
        self.set_header("Content-Type", "application/json")
        paths = [
            entry["path"]
            for entry in foldersAtPath(os.path.dirname(xhtml_unescape(term)), includeFiles=bool(int(includeFiles)), fileTypes=fileTypes.split(","))
            if "path" in entry
        ]

        return json.dumps(paths)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2017-8-30

@author: generated by @lolobosse script
'''
LOCALE = [
    ["juuri äsken", "juuri nyt"],
    ["%s sekuntia sitten", "%s sekunnin päästä"],
    ["minuutti sitten", "minuutin päästä"],
    ["%s minuuttia sitten", "%s minuutin päästä"],
    ["tunti sitten", "tunnin päästä"],
    ["%s tuntia sitten", "%s tunnin päästä"],
    ["päivä sitten", "päivän päästä"],
    ["%s päivää sitten", "%s päivän päästä"],
    ["viikko sitten", "viikon päästä"],
    ["%s viikkoa sitten", "%s viikon päästä"],
    ["kuukausi sitten", "kuukauden päästä"],
    ["%s kuukautta sitten", "%s kuukauden päästä"],
    ["vuosi sitten", "vuoden päästä"],
    ["%s vuotta sitten", "%s vuoden päästä"]
]
        


"""
Test NZBProvider
"""

import os
import unittest

from sickchill import settings
from sickchill.providers.GenericProvider import GenericProvider
from sickchill.providers.nzb.NZBProvider import NZBProvider

from .test_generic_provider import GenericProviderTests


class NZBProviderTests(GenericProviderTests):
    """
    Test NZBProvider
    """

    def test___init__(self):
        """
        Test __init__
        """
        assert NZBProvider("Test Provider").provider_type == GenericProvider.NZB

    def test_is_active(self):
        """
        Test is_active
        """
        test_cases = {
            (False, False): False,
            (False, None): False,
            (False, True): False,
            (None, False): False,
            (None, None): False,
            (None, True): False,
            (True, False): False,
            (True, None): False,
            (True, True): True,
        }

        for ((use_nzb, enabled), result) in test_cases.items():
            settings.USE_NZBS = use_nzb

            provider = NZBProvider("Test Provider")
            provider.enabled = enabled

            assert provider.is_active == result

    def test__get_size(self):
        """
        Test _get_size
        """
        items_list = [
            None,
            {},
            {"links": None},
            {"links": []},
            {"links": [{}]},
            {"links": [{"length": 1}, {"length": None}, {"length": 3}]},
            {"links": [{"length": 1}, {"length": ""}, {"length": 3}]},
            {"links": [{"length": 1}, {"length": "0"}, {"length": 3}]},
            {"links": [{"length": 1}, {"length": "123"}, {"length": 3}]},
            {"links": [{"length": 1}, {"length": "12.3"}, {"length": 3}]},
            {"links": [{"length": 1}, {"length": "-123"}, {"length": 3}]},
            {"links": [{"length": 1}, {"length": "-12.3"}, {"length": 3}]},
            {"links": [{"length": 1}, {"length": 0}, {"length": 3}]},
            {"links": [{"length": 1}, {"length": 123}, {"length": 3}]},
            {"links": [{"length": 1}, {"length": 12.3}, {"length": 3}]},
            {"links": [{"length": 1}, {"length": -123}, {"length": 3}]},
            {"links": [{"length": 1}, {"length": -12.3}, {"length": 3}]},
        ]
        results_list = [-1, -1, -1, -1, -1, -1, -1, 0, 123, -1, -123, -1, 0, 123, 12, -123, -12]

        unicode_items_list = [
            {"links": None},
            {"links": []},
            {"links": [{}]},
            {"links": [{"length": 1}, {"length": None}, {"length": 3}]},
            {"links": [{"length": 1}, {"length": ""}, {"length": 3}]},
            {"links": [{"length": 1}, {"length": "0"}, {"length": 3}]},
            {"links": [{"length": 1}, {"length": "123"}, {"length": 3}]},
            {"links": [{"length": 1}, {"length": "12.3"}, {"length": 3}]},
            {"links": [{"length": 1}, {"length": "-123"}, {"length": 3}]},
            {"links": [{"length": 1}, {"length": "-12.3"}, {"length": 3}]},
            {"links": [{"length": 1}, {"length": 0}, {"length": 3}]},
            {"links": [{"length": 1}, {"length": 123}, {"length": 3}]},
            {"links": [{"length": 1}, {"length": 12.3}, {"length": 3}]},
            {"links": [{"length": 1}, {"length": -123}, {"length": 3}]},
            {"links": [{"length": 1}, {"length": -12.3}, {"length": 3}]},
        ]
        unicode_results_list = [-1, -1, -1, -1, -1, 0, 123, -1, -123, -1, 0, 123, 12, -123, -12]

        assert len(items_list) == len(results_list), "Number of parameters ({0:d}) and results ({1:d}) does not match".format(
            len(items_list), len(results_list))

        assert len(unicode_items_list) == len(unicode_results_list), "Number of parameters ({0:d}) and results ({1:d}) does not match".format(
            len(unicode_items_list), len(unicode_results_list))

        for (index, item) in enumerate(items_list):
            assert NZBProvider("Test Provider")._get_size(item) == results_list[index]

        for (index, item) in enumerate(unicode_items_list):
            assert NZBProvider("Test Provider")._get_size(item) == unicode_results_list[index]

    def test__get_storage_dir(self):
        """
        Test _get_storage_dir
        """
        test_cases = [None, 123, 12.3, "", os.path.join("some", "path", "to", "folder")]

        for nzb_dir in test_cases:
            settings.NZB_DIR = nzb_dir

            assert NZBProvider("Test Provider")._get_storage_dir() == nzb_dir


if __name__ == "__main__":
    print("=====> Testing {0}".format(__file__))

    SUITE = unittest.TestLoader().loadTestsFromTestCase(NZBProviderTests)
    unittest.TextTestRunner(verbosity=2).run(SUITE)



from sickchill.oldbeard import db


# Add new migrations at the bottom of the list; subclass the previous migration.
class InitialSchema(db.SchemaUpgrade):
    def test(self):
        return self.has_table("db_version")

    def execute(self):
        queries = [
            ('CREATE TABLE failed ("release" TEXT, size NUMERIC, provider TEXT);',),
            (
                'CREATE TABLE history (date NUMERIC, size NUMERIC, "release" TEXT, provider TEXT, old_status NUMERIC DEFAULT 0, '
                "showid NUMERIC DEFAULT -1, season NUMERIC DEFAULT -1, episode NUMERIC DEFAULT -1);",
            ),
            ("CREATE TABLE db_version (db_version INTEGER);",),
            ("INSERT INTO db_version (db_version) VALUES (1);",),
        ]
        for query in queries:
            if len(query) == 1:
                self.connection.action(query[0])
            else:
                self.connection.action(query[0], query[1:])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
API functions that can be used by external software
"""

try:
    from collections import OrderedDict
except ImportError:  # pragma: no-cover
    from ordereddict import OrderedDict  # pylint:disable=import-error

import os
import traceback

import six
from rebulk.introspector import introspect

from .__version__ import __version__
from .options import parse_options, load_config, merge_options
from .rules import rebulk_builder


class GuessitException(Exception):
    """
    Exception raised when guessit fails to perform a guess because of an internal error.
    """

    def __init__(self, string, options):
        super(GuessitException, self).__init__("An internal error has occured in guessit.\n"
                                               "===================== Guessit Exception Report =====================\n"
                                               "version=%s\n"
                                               "string=%s\n"
                                               "options=%s\n"
                                               "--------------------------------------------------------------------\n"
                                               "%s"
                                               "--------------------------------------------------------------------\n"
                                               "Please report at "
                                               "https://github.com/guessit-io/guessit/issues.\n"
                                               "====================================================================" %
                                               (__version__, str(string), str(options), traceback.format_exc()))

        self.string = string
        self.options = options


def configure(options=None, rules_builder=rebulk_builder, force=False):
    """
    Load configuration files and initialize rebulk rules if required.

    :param options:
    :type options: dict
    :param rules_builder:
    :type rules_builder:
    :param force:
    :type force: bool
    :return:
    """
    default_api.configure(options, rules_builder=rules_builder, force=force)


def guessit(string, options=None):
    """
    Retrieves all matches from string as a dict
    :param string: the filename or release name
    :type string: str
    :param options:
    :type options: str|dict
    :return:
    :rtype:
    """
    return default_api.guessit(string, options)


def properties(options=None):
    """
    Retrieves all properties with possible values that can be guessed
    :param options:
    :type options: str|dict
    :return:
    :rtype:
    """
    return default_api.properties(options)


def suggested_expected(titles, options=None):
    """
    Return a list of suggested titles to be used as `expected_title` based on the list of titles
    :param titles: the filename or release name
    :type titles: list|set|dict
    :param options:
    :type options: str|dict
    :return:
    :rtype: list of str
    """
    return default_api.suggested_expected(titles, options)


class GuessItApi(object):
    """
    An api class that can be configured with custom Rebulk configuration.
    """

    def __init__(self):
        """Default constructor."""
        self.rebulk = None
        self.config = None
        self.load_config_options = None
        self.advanced_config = None

    @classmethod
    def _fix_encoding(cls, value):
        if isinstance(value, list):
            return [cls._fix_encoding(item) for item in value]
        if isinstance(value, dict):
            return {cls._fix_encoding(k): cls._fix_encoding(v) for k, v in value.items()}
        if six.PY2 and isinstance(value, six.text_type):
            return value.encode('utf-8')
        if six.PY3 and isinstance(value, six.binary_type):
            return value.decode('ascii')
        return value

    @classmethod
    def _has_same_properties(cls, dic1, dic2, values):
        for value in values:
            if dic1.get(value) != dic2.get(value):
                return False
        return True

    def configure(self, options=None, rules_builder=rebulk_builder, force=False, sanitize_options=True):
        """
        Load configuration files and initialize rebulk rules if required.

        :param options:
        :type options: str|dict
        :param rules_builder:
        :type rules_builder:
        :param force:
        :type force: bool
        :return:
        :rtype: dict
        """
        if sanitize_options:
            options = parse_options(options, True)
            options = self._fix_encoding(options)

        if self.config is None or self.load_config_options is None or force or \
                not self._has_same_properties(self.load_config_options,
                                              options,
                                              ['config', 'no_user_config', 'no_default_config']):
            config = load_config(options)
            config = self._fix_encoding(config)
            self.load_config_options = options
        else:
            config = self.config

        advanced_config = merge_options(config.get('advanced_config'), options.get('advanced_config'))

        should_build_rebulk = force or not self.rebulk or not self.advanced_config or \
                              self.advanced_config != advanced_config

        if should_build_rebulk:
            self.advanced_config = advanced_config
            self.rebulk = rules_builder(advanced_config)

        self.config = config
        return self.config

    def guessit(self, string, options=None):  # pylint: disable=too-many-branches
        """
        Retrieves all matches from string as a dict
        :param string: the filename or release name
        :type string: str|Path
        :param options:
        :type options: str|dict
        :return:
        :rtype:
        """
        try:
            from pathlib import Path
            if isinstance(string, Path):
                try:
                    # Handle path-like object
                    string = os.fspath(string)
                except AttributeError:
                    string = str(string)
        except ImportError:
            pass

        try:
            options = parse_options(options, True)
            options = self._fix_encoding(options)
            config = self.configure(options, sanitize_options=False)
            options = merge_options(config, options)
            result_decode = False
            result_encode = False

            if six.PY2:
                if isinstance(string, six.text_type):
                    string = string.encode("utf-8")
                    result_decode = True
                elif isinstance(string, six.binary_type):
                    string = six.binary_type(string)
            if six.PY3:
                if isinstance(string, six.binary_type):
                    string = string.decode('ascii')
                    result_encode = True
                elif isinstance(string, six.text_type):
                    string = six.text_type(string)

            matches = self.rebulk.matches(string, options)
            if result_decode:
                for match in matches:
                    if isinstance(match.value, six.binary_type):
                        match.value = match.value.decode("utf-8")
            if result_encode:
                for match in matches:
                    if isinstance(match.value, six.text_type):
                        match.value = match.value.encode("ascii")
            return matches.to_dict(options.get('advanced', False), options.get('single_value', False),
                                   options.get('enforce_list', False))
        except:
            raise GuessitException(string, options)

    def properties(self, options=None):
        """
        Grab properties and values that can be generated.
        :param options:
        :type options:
        :return:
        :rtype:
        """
        options = parse_options(options, True)
        options = self._fix_encoding(options)
        config = self.configure(options, sanitize_options=False)
        options = merge_options(config, options)
        unordered = introspect(self.rebulk, options).properties
        ordered = OrderedDict()
        for k in sorted(unordered.keys(), key=six.text_type):
            ordered[k] = list(sorted(unordered[k], key=six.text_type))
        if hasattr(self.rebulk, 'customize_properties'):
            ordered = self.rebulk.customize_properties(ordered)
        return ordered

    def suggested_expected(self, titles, options=None):
        """
        Return a list of suggested titles to be used as `expected_title` based on the list of titles
        :param titles: the filename or release name
        :type titles: list|set|dict
        :param options:
        :type options: str|dict
        :return:
        :rtype: list of str
        """
        suggested = []
        for title in titles:
            guess = self.guessit(title, options)
            if len(guess) != 2 or 'title' not in guess:
                suggested.append(title)

        return suggested


default_api = GuessItApi()



import fanart
from fanart.items import LeafItem, Immutable, ResourceItem
__all__ = (
    'CharacterItem',
    'ArtItem',
    'LogoItem',
    'BackgroundItem',
    'SeasonItem',
    'SeasonPosterItem',
    'SeasonBannerItem',
    'ThumbItem',
    'HdLogoItem',
    'HdArtItem',
    'PosterItem',
    'BannerItem',
    'TvShow',
)


class TvItem(LeafItem):
    @Immutable.mutablemethod
    def __init__(self, id, url, likes, lang):
        super(TvItem, self).__init__(id, url, likes)
        self.lang = lang


class SeasonedTvItem(TvItem):
    @Immutable.mutablemethod
    def __init__(self, id, url, likes, lang, season):
        super(SeasonedTvItem, self).__init__(id, url, likes, lang)
        self.season = 0 if season == 'all' else int(season or 0)


class CharacterItem(TvItem):
    KEY = fanart.TYPE.TV.CHARACTER


class ArtItem(TvItem):
    KEY = fanart.TYPE.TV.ART


class LogoItem(TvItem):
    KEY = fanart.TYPE.TV.LOGO


class BackgroundItem(SeasonedTvItem):
    KEY = fanart.TYPE.TV.BACKGROUND


class SeasonItem(SeasonedTvItem):
    KEY = fanart.TYPE.TV.SEASONTHUMB


class SeasonPosterItem(SeasonedTvItem):
    KEY = fanart.TYPE.TV.SEASONPOSTER


class SeasonBannerItem(SeasonedTvItem):
    KEY = fanart.TYPE.TV.SEASONBANNER


class ThumbItem(TvItem):
    KEY = fanart.TYPE.TV.THUMB


class HdLogoItem(TvItem):
    KEY = fanart.TYPE.TV.HDLOGO


class HdArtItem(TvItem):
    KEY = fanart.TYPE.TV.HDART


class PosterItem(TvItem):
    KEY = fanart.TYPE.TV.POSTER


class BannerItem(TvItem):
    KEY = fanart.TYPE.TV.BANNER


class TvShow(ResourceItem):
    WS = fanart.WS.TV

    @Immutable.mutablemethod
    def __init__(
            self, name, tvdbid, backgrounds, characters, arts, logos,
            seasons, thumbs, hdlogos, hdarts, posters, banners,
            season_posters, season_banners):
        self.name = name
        self.tvdbid = tvdbid
        self.backgrounds = backgrounds
        self.characters = characters
        self.arts = arts
        self.logos = logos
        self.seasons = seasons
        self.thumbs = thumbs
        self.hdlogos = hdlogos
        self.hdarts = hdarts
        self.posters = posters
        self.banners = banners
        self.season_posters = posters
        self.season_banners = banners

    @classmethod
    def from_dict(cls, resource):
        minimal_keys = {'name', 'thetvdb_id'}
        assert all(k in resource for k in minimal_keys), 'Bad Format Map'
        return cls(
            name=resource['name'],
            tvdbid=resource['thetvdb_id'],
            backgrounds=BackgroundItem.extract(resource),
            characters=CharacterItem.extract(resource),
            arts=ArtItem.extract(resource),
            logos=LogoItem.extract(resource),
            seasons=SeasonItem.extract(resource),
            thumbs=ThumbItem.extract(resource),
            hdlogos=HdLogoItem.extract(resource),
            hdarts=HdArtItem.extract(resource),
            posters=PosterItem.extract(resource),
            banners=BannerItem.extract(resource),
            season_posters=SeasonPosterItem.extract(resource),
            season_banners=SeasonBannerItem.extract(resource),
        )



import re
import traceback
import urllib
from collections import OrderedDict
from urllib.parse import urljoin

import validators

from sickchill import logger
from sickchill.helper.common import convert_size, try_int
from sickchill.oldbeard import tvcache
from sickchill.oldbeard.bs4_parser import BS4Parser
from sickchill.providers.torrent.TorrentProvider import TorrentProvider


class Provider(TorrentProvider):
    def __init__(self):

        super().__init__("KickAssTorrents")

        self.public = True

        self.confirmed = True
        self.minseed = 0
        self.minleech = 0
        self.confirmed = True

        self.mirrors = []
        self.disabled_mirrors = []

        # https://kickasskat.org/tv?field=time_add&sorder=desc
        # https://kickasskat.org/usearch/{query}/?category=tv&field=seeders&sorder=desc

        self.url = "https://kickasskat.org"
        self.urls = None

        self.custom_url = None

        self.cache = tvcache.TVCache(self)

        self.rows_selector = dict(class_=re.compile(r"even|odd"), id=re.compile(r"torrent_.*_torrents"))

    def search(self, search_strings, age=0, ep_obj=None):
        results = []
        if not (self.url and self.urls):
            self.find_domain()
            if not (self.url and self.urls):
                return results

        anime = (self.show and self.show.anime) or (ep_obj and ep_obj.show and ep_obj.show.anime) or False
        search_params = {"field": "seeders", "sorder": "desc", "category": ("tv", "anime")[anime]}

        for mode in search_strings:
            items = []
            logger.debug(_("Search Mode: {mode}".format(mode=mode)))
            for search_string in {*search_strings[mode]}:

                # search_params["q"] = (search_string, None)[mode == "RSS"]
                search_params["field"] = ("seeders", "time_add")[mode == "RSS"]

                if mode != "RSS":
                    if anime:
                        continue

                    logger.debug(_("Search String: {search_string}".format(search_string=search_string)))

                    search_url = self.urls["search"].format(q=search_string)
                else:
                    search_url = self.urls["rss"]

                if self.custom_url:
                    if not validators.url(self.custom_url):
                        logger.warning("Invalid custom url: {0}".format(self.custom_url))
                        return results
                    search_url = urljoin(self.custom_url, search_url.split(self.url)[1])

                data = self.get_url(search_url, params=OrderedDict(sorted(list(search_params.items()), key=lambda x: x[0])), returns="text")
                if not data:
                    logger.info("{url} did not return any data, it may be disabled. Trying to get a new domain".format(url=self.url))
                    self.disabled_mirrors.append(self.url)
                    self.find_domain()
                    if self.url in self.disabled_mirrors:
                        logger.info("Could not find a better mirror to try.")
                        logger.info("The search did not return data, if the results are on the site maybe try a custom url, or a different one")
                        return results

                    # This will recurse a few times until all of the mirrors are exhausted if none of them work.
                    return self.search(search_strings, age, ep_obj)

                with BS4Parser(data, "html5lib") as html:
                    labels = [cell.get_text() for cell in html.find(class_="firstr")("th")]
                    logger.info("Found {} results".format(len(html("tr", **self.rows_selector))))
                    for result in html("tr", **self.rows_selector):
                        try:
                            download_url = urllib.parse.unquote_plus(result.find(title="Torrent magnet link")["href"].split("url=")[1]) + self._custom_trackers
                            parsed_magnet = urllib.parse.parse_qs(download_url)
                            torrent_hash = self.hash_from_magnet(download_url)
                            title = result.find(class_="torrentname").find(class_="cellMainLink").get_text(strip=True)
                            if title.endswith("..."):
                                title = parsed_magnet["dn"][0]

                            if not (title and download_url):
                                if mode != "RSS":
                                    logger.debug("Discarding torrent because We could not parse the title and url")
                                continue

                            seeders = try_int(result.find(class_="green").get_text(strip=True))
                            leechers = try_int(result.find(class_="red").get_text(strip=True))

                            # Filter unseeded torrent
                            if seeders < self.minseed or leechers < self.minleech:
                                if mode != "RSS":
                                    logger.debug(
                                        "Discarding torrent because it doesn't meet the minimum seeders or leechers: {0} (S:{1} L:{2})".format(
                                            title, seeders, leechers
                                        )
                                    )
                                continue

                            if self.confirmed and not result.find(class_="ka-green"):
                                if mode != "RSS":
                                    logger.debug("Found result " + title + " but that doesn't seem like a verified result so I'm ignoring it")
                                continue

                            torrent_size = result("td")[labels.index("size")].get_text(strip=True)
                            size = convert_size(torrent_size) or -1

                            item = {"title": title, "link": download_url, "size": size, "seeders": seeders, "leechers": leechers, "hash": torrent_hash}
                            if mode != "RSS":
                                logger.debug("Found result: {0} with {1} seeders and {2} leechers".format(title, seeders, leechers))

                            items.append(item)

                        except (AttributeError, TypeError, KeyError, ValueError, Exception):
                            logger.info(traceback.format_exc())
                            continue

            # For each search mode sort all the items by seeders if available
            items.sort(key=lambda d: try_int(d.get("seeders", 0)), reverse=True)

            results += items

        return results

    def find_domain(self):
        data = self.get_url("https://kickass2.help")
        if data:
            with BS4Parser(data, "html5lib") as html:
                mirrors = html(class_="domainLink")
                if mirrors:
                    self.mirrors = []
                for mirror in mirrors:
                    domain = mirror["href"]
                    if domain not in self.disabled_mirrors:
                        self.mirrors.append(mirror["href"])

        if self.mirrors:
            self.url = self.mirrors[0]
            logger.info("Setting mirror to use to {url}".format(url=self.url))
        else:
            logger.warning(
                "Unable to get a working mirror for kickasstorrents, you might need to enable another provider and disable KAT until KAT starts working "
                "again."
            )

        self.urls = {"search": urljoin(self.url, "/usearch/{q}/"), "rss": urljoin(self.url, "/tv/")}

        return self.url



import locale
import logging
import logging.handlers
import os
import platform
import re
import sys
import traceback
from logging import NullHandler
from urllib.parse import quote

from sickchill.init_helpers import setup_lib_path

setup_lib_path()

from github import InputFileContent
from github.GithubException import RateLimitExceededException, TwoFactorException

from sickchill import settings
from sickchill.helper.common import dateTimeFormat
from sickchill.oldbeard import classes

# log levels
ERROR = logging.ERROR
WARNING = logging.WARNING
INFO = logging.INFO
DEBUG = logging.DEBUG
DB = 5

LOGGING_LEVELS = {
    "ERROR": ERROR,
    "WARNING": WARNING,
    "INFO": INFO,
    "DEBUG": DEBUG,
    "DB": DB,
}

censored_items = {}


class DispatchFormatter(logging.Formatter, object):
    """
    Censor information such as API keys, user names, and passwords from the Log
    """

    def __init__(self, fmt=None, datefmt=None, style="{"):
        super().__init__(fmt, datefmt, style=style)

    def format(self, record):
        """
        Strips censored items from string and formats the log line

        :param record: to format
        """

        msg = record.msg

        # set of censored items
        censored = {item for _, item in censored_items.items() if item}
        # set of censored items and urlencoded counterparts
        censored = list(censored | {quote(item) for item in censored})
        # sort the list in order of descending length so that entire item is censored
        # e.g. password and password_1 both get censored instead of getting ********_1
        censored.sort(key=len, reverse=True)

        if not isinstance(msg, (str, bytes)):
            msg = repr(msg)

        for item in censored:
            try:
                # passwords that include ++ for example will error. Cannot escape or it wont match at all.
                msg = re.sub(fr"\b({item})\b", "*" * 8, msg)
            except re.error:
                msg = msg.replace(item, "*" * 8)
            except TypeError:
                print(msg)

        # Needed because Newznab apikey isn't stored as key=value in a section.
        msg = re.sub(r"([&?]r|[&?]apikey|[&?]jackett_apikey|[&?]api_key)(?:=|%3D)[^&]*([&\w]?)", r"\1=**********\2", msg, re.I)

        if record.levelno == ERROR:
            classes.ErrorViewer.add(classes.UIError(msg))
        elif record.levelno == WARNING:
            classes.WarningViewer.add(classes.UIError(msg))

        return super().format(record)


class Logger(object):
    """
    Logger to create log entries
    """

    def __init__(self):
        self.logger = logging.getLogger("sickchill")

        self.loggers = [
            logging.getLogger("sickchill"),
            logging.getLogger("sickchill.movie"),
            logging.getLogger("tornado.general"),
            logging.getLogger("tornado.application"),
            logging.getLogger("subliminal"),
            logging.getLogger("tornado.access"),
            logging.getLogger("imdbpy.parser.http.piculet"),
        ]

        self.console_logging = False
        self.file_logging = False
        self.debug_logging = False
        self.database_logging = False
        self.log_file = None

        self.submitter_running = False

    def init_logging(self, console_logging=False, file_logging=False, debug_logging=False, database_logging=False):
        """
        Initialize logging

        :param console_logging: True if logging to console
        :param file_logging: True if logging to file
        :param debug_logging: True if debug logging is enabled
        :param database_logging: True if logging database access
        """
        self.log_file = self.log_file or os.path.join(settings.LOG_DIR, "sickchill.log")

        global log_file
        log_file = self.log_file

        self.debug_logging = debug_logging
        self.console_logging = console_logging
        self.file_logging = file_logging
        self.database_logging = database_logging

        logging.addLevelName(DB, "DB")  # add a new logging level DB
        logging.getLogger().addHandler(NullHandler())  # nullify root logger

        # set custom root logger
        for logger in self.loggers:
            if logger is not self.logger:
                logger.root = self.logger
                logger.parent = self.logger

        log_level = DB if self.database_logging else DEBUG if self.debug_logging else INFO

        # set minimum logging level allowed for loggers
        for logger in self.loggers:
            if logger.name in ("subliminal", "tornado.access", "tornado.general", "imdbpy.parser.http.piculet"):
                logger.setLevel("CRITICAL")
            else:
                logger.setLevel(log_level)

        log_format = "{asctime} {levelname} :: {threadName} :: {message}"
        # console log handler
        if self.console_logging:
            console = logging.StreamHandler()
            console.setFormatter(DispatchFormatter(log_format, dateTimeFormat))
            console.setLevel(log_level)

            for logger in self.loggers:
                logger.addHandler(console)

        # rotating log file handler
        if self.file_logging:
            rfh = logging.handlers.RotatingFileHandler(self.log_file, maxBytes=int(settings.LOG_SIZE * 1048576), backupCount=settings.LOG_NR)
            rfh.setFormatter(DispatchFormatter(log_format, dateTimeFormat))
            rfh.setLevel(log_level)

            for logger in self.loggers:
                logger.addHandler(rfh)

    def set_level(self):
        """
        Sets the logging level for root and all of our loggers
        """
        self.debug_logging = settings.DEBUG
        self.database_logging = settings.DBDEBUG

        level = DB if self.database_logging else DEBUG if self.debug_logging else INFO
        for logger in self.loggers:
            if logger.name in ("subliminal", "tornado.access", "tornado.general"):
                logger.setLevel("CRITICAL")
            else:
                logger.setLevel(level)
                for handler in logger.handlers:
                    handler.setLevel(level)

    @staticmethod
    def shutdown():
        """
        Shut down the logger
        """
        logging.shutdown()

    def log_error_and_exit(self, error_msg, *args, **kwargs):
        self.logger.log(ERROR, error_msg, *args, **kwargs)

        if not self.console_logging:
            sys.exit(error_msg)
        else:
            sys.exit(1)

    def submit_errors(self):

        submitter_result = ""
        issue_id = None

        if not all((settings.GIT_TOKEN, settings.DEBUG, settings.gh, classes.ErrorViewer.errors)):
            submitter_result = "Please set your GitHub token in the config and enable debug. Unable to submit issue ticket to GitHub!"
            return submitter_result, issue_id

        try:
            from sickchill.update_manager import UpdateManager

            update_manager = UpdateManager()
            update_manager.check_for_new_version()
            commits_behind = update_manager.get_num_commits_behind()
        except Exception:
            submitter_result = "Could not check if your SickChill is updated, unable to submit issue ticket to GitHub!"
            return submitter_result, issue_id

        if commits_behind is None or commits_behind > 0:
            submitter_result = "Please update SickChill, unable to submit issue ticket to GitHub with an outdated version!"
            return submitter_result, issue_id

        if self.submitter_running:
            submitter_result = "Issue submitter is running, please wait for it to complete"
            return submitter_result, issue_id

        self.submitter_running = True

        try:
            # read log file
            __log_data = None

            if os.path.isfile(self.log_file):
                with open(self.log_file) as log_f:
                    __log_data = log_f.readlines()

            for i in range(1, int(settings.LOG_NR)):
                f_name = f"{self.log_file}.{i}"
                if os.path.isfile(f_name) and (len(__log_data) <= 500):
                    with open(f_name) as log_f:
                        __log_data += log_f.readlines()

            __log_data = list(reversed(__log_data))

            # parse and submit errors to issue tracker
            for cur_error in sorted(classes.ErrorViewer.errors, key=lambda error: error.time, reverse=True)[:500]:
                try:
                    title_error = str(cur_error.title)
                    if not title_error or title_error == "None":
                        title_error = re.match(r"^[A-Za-z0-9\-\[\] :]+::\s(?:\[[\w]{7}\])\s*(.*)$", cur_error.message).group(1)

                    if len(title_error) > 1000:
                        title_error = title_error[0:1000]

                except Exception as err_msg:
                    self.logger.log(ERROR, f"Unable to get error title : {err_msg}")
                    title_error = "UNKNOWN"

                gist = None
                regex = fr"^(?P<time>{re.escape(cur_error.time)})\s+(?P<level>[A-Z]+)\s+[A-Za-z0-9\-\[\] :]+::.*$"
                for i, data in enumerate(__log_data):
                    match = re.match(regex, data)
                    if match:
                        level = match.group("level")
                        if LOGGING_LEVELS[level] == ERROR:
                            paste_data = "".join(__log_data[i : i + 50])
                            if paste_data:
                                gist = settings.gh.get_user().create_gist(False, {"sickchill.log": InputFileContent(paste_data)})
                            break
                    else:
                        gist = "No ERROR found"

                try:
                    locale_name = locale.getdefaultlocale()[1]
                except Exception:
                    locale_name = "unknown"

                if gist and gist != "No ERROR found":
                    log_link = f"Link to Log: {gist.html_url}"
                else:
                    log_link = "No Log available with ERRORS:"

                msg = [
                    "### INFO",
                    f"Python Version: **{sys.version[:120]}**".replace("\n", ""),
                    f"Operating System: **{platform.platform()}**",
                    f"Locale: {locale_name}",
                    f"Branch: **{settings.BRANCH}**",
                    f"Commit: SickChill/SickChill@{settings.CUR_COMMIT_HASH}",
                    log_link,
                    "### ERROR",
                    "```",
                    cur_error.message,
                    "```",
                    "---",
                    "_STAFF NOTIFIED_: @SickChill/owners @SickChill/moderators",
                ]

                message = "\n".join(msg)
                title_error = f"[APP SUBMITTED]: {title_error}"

                repo = settings.gh.get_organization(settings.GIT_ORG).get_repo(settings.GIT_REPO)
                reports = repo.get_issues(state="all")

                def is_ascii_error(title):
                    # [APP SUBMITTED]: 'ascii' codec can't encode characters in position 00-00: ordinal not in range(128)
                    # [APP SUBMITTED]: 'charmap' codec can't decode byte 0x00 in position 00: character maps to <undefined>
                    return re.search(r".* codec can\'t .*code .* in position .*:", title) is not None

                def is_malformed_error(title):
                    # [APP SUBMITTED]: not well-formed (invalid token): line 0, column 0
                    return re.search(r".* not well-formed \(invalid token\): line .* column .*", title) is not None

                ascii_error = is_ascii_error(title_error)
                malformed_error = is_malformed_error(title_error)

                issue_found = False
                for report in reports:
                    if (
                        title_error.rsplit(" :: ")[-1] in report.title
                        or (malformed_error and is_malformed_error(report.title))
                        or (ascii_error and is_ascii_error(report.title))
                    ):

                        issue_id = report.number
                        if not report.raw_data["locked"]:
                            if report.create_comment(message):
                                submitter_result = f"Commented on existing issue #{issue_id} successfully!"
                            else:
                                submitter_result = f"Failed to comment on found issue #{issue_id}!"
                        else:
                            submitter_result = f"Issue #{issue_id} is locked, check GitHub to find info about the error."

                        issue_found = True
                        break

                if not issue_found:
                    issue = repo.create_issue(title_error, message)
                    if issue:
                        issue_id = issue.number
                        submitter_result = f"Your issue ticket #{issue_id} was submitted successfully!"
                    else:
                        submitter_result = "Failed to create a new issue!"

                if issue_id and cur_error in classes.ErrorViewer.errors:
                    # clear error from error list
                    classes.ErrorViewer.errors.remove(cur_error)
        except RateLimitExceededException:
            submitter_result = "Your Github user has exceeded its API rate limit, please try again later"
            issue_id = None
        except TwoFactorException:
            submitter_result = "Your Github account requires Two-Factor Authentication, " "please change your auth method in the config"
            issue_id = None
        except Exception:
            self.logger.log(ERROR, traceback.format_exc())
            submitter_result = "Exception generated in issue submitter, please check the log"
            issue_id = None
        finally:
            self.submitter_running = False

        return submitter_result, issue_id


class Wrapper(object):
    instance = Logger()

    def __init__(self, wrapped):
        self.wrapped = wrapped

    def __getattr__(self, name):
        try:
            return getattr(self.wrapped, name)
        except AttributeError:
            return getattr(self.instance, name)


_globals = sys.modules[__name__] = Wrapper(sys.modules[__name__])

LOG_FILTERS = {
    "<NONE>": _("&lt;No Filter&gt;"),
    "DAILYSEARCHER": _("Daily Searcher"),
    "BACKLOG": _("Backlog"),
    "SHOWUPDATER": _("Show Updater"),
    "CHECKVERSION": _("Check Version"),
    "SHOWQUEUE": _("Show Queue"),
    "SEARCHQUEUE": _("Search Queue (All)"),
    "SEARCHQUEUE-DAILY-SEARCH": _("Search Queue (Daily Searcher)"),
    "SEARCHQUEUE-BACKLOG": _("Search Queue (Backlog)"),
    "SEARCHQUEUE-MANUAL": _("Search Queue (Manual)"),
    "SEARCHQUEUE-RETRY": _("Search Queue (Retry/Failed)"),
    "SEARCHQUEUE-RSS": _("Search Queue (RSS)"),
    "FINDPROPERS": _("Find Propers"),
    "POSTPROCESSOR": _("Postprocessor"),
    "FINDSUBTITLES": _("Find Subtitles"),
    "TRAKTCHECKER": _("Trakt Checker"),
    "EVENT": _("Event"),
    "ERROR": _("Error"),
    "TORNADO": _("Tornado"),
    "Thread": _("Thread"),
    "MAIN": _("Main"),
}


def log_data(min_level, log_filter, log_search, max_lines):
    regex = r"^\d{4}\-\d{2}\-\d{2} \d{2}\:\d{2}:\d{2} ([A-Z]+) \:\: (.+?) \:\: (.*)$"
    if log_filter not in LOG_FILTERS:
        log_filter = "<NONE>"

    final_data = []

    log_files = []
    if os.path.isfile(Wrapper.instance.log_file):
        log_files.append(Wrapper.instance.log_file)

        for i in range(1, int(settings.LOG_NR)):
            name = Wrapper.instance.log_file + "." + str(i)
            if not os.path.isfile(name):
                break
            log_files.append(name)
    else:
        return final_data

    data = []
    for _log_file in log_files:
        if len(data) < max_lines:
            with open(_log_file, "r") as f:
                data += [line.strip() + "\n" for line in reversed(f.readlines()) if line.strip()]
        else:
            break

    found_lines = 0
    for x in data:
        match = re.match(regex, x)

        if match:
            level = match.group(1)
            log_name = match.group(2)

            if not settings.DEBUG and level == "DEBUG":
                continue

            if not settings.DBDEBUG and level == "DB":
                continue

            if level not in LOGGING_LEVELS:
                final_data.append("AA " + x)
                found_lines += 1
            elif log_search and log_search.lower() in x.lower():
                final_data.append(x)
                found_lines += 1
            elif not log_search and LOGGING_LEVELS[level] >= int(min_level) and (log_filter == "<NONE>" or log_name.startswith(log_filter)):
                final_data.append(x)
                found_lines += 1
        else:
            final_data.append("AA " + x)
            found_lines += 1

        if found_lines >= max_lines:
            break

    return final_data


log_error_and_exit = Wrapper.instance.log_error_and_exit
log_file = Wrapper.instance.log_file

log = Wrapper.instance.logger.log
debug = Wrapper.instance.logger.debug
info = Wrapper.instance.logger.info
warning = Wrapper.instance.logger.warning
error = Wrapper.instance.logger.error
exception = logging.exception
critical = Wrapper.instance.logger.critical


def database(msg, *args, **kwargs):
    """
    Handler to add database sql logging
    @param msg: message to be logged
    @param args: arguments to logger
    @param kwargs:  kwargs to logger
    """
    if settings.DBDEBUG:
        debug(msg, args, kwargs)


submit_errors = Wrapper.instance.submit_errors
init_logging = Wrapper.instance.init_logging
set_level = Wrapper.instance.set_level
shutdown = Wrapper.instance.shutdown



from sickchill import logger, settings
from sickchill.adba.aniDBerrors import AniDBCommandTimeoutError

from . import db, helpers


class BlackAndWhiteList(object):
    blacklist = []
    whitelist = []

    def __init__(self, show_id):
        if not show_id:
            raise BlackWhitelistNoShowIDException()
        self.show_id = show_id
        self.load()

    def load(self):
        """
        Builds black and whitelist
        """
        logger.debug("Building black and white list for {id}".format(id=self.show_id))
        self.blacklist = self._load_list("blacklist")
        self.whitelist = self._load_list("whitelist")

    def _add_keywords(self, table, values):
        """
        DB: Adds keywords into database for current show

        :param table: SQL table to add keywords to
        :param values: Values to be inserted in table
        """
        main_db_con = db.DBConnection()
        for value in values:
            main_db_con.action("INSERT INTO [" + table + "] (show_id, keyword) VALUES (?,?)", [self.show_id, value])

    def set_black_keywords(self, values):
        """
        Sets blacklist to new value

        :param values: Complete list of keywords to be set as blacklist
        """
        self._del_all_keywords("blacklist")
        self._add_keywords("blacklist", values)
        self.blacklist = values
        logger.debug("Blacklist set to: {blacklist}".format(blacklist=self.blacklist))

    def set_white_keywords(self, values):
        """
        Sets whitelist to new value

        :param values: Complete list of keywords to be set as whitelist
        """
        self._del_all_keywords("whitelist")
        self._add_keywords("whitelist", values)
        self.whitelist = values
        logger.debug("Whitelist set to: {whitelist}".format(whitelist=self.whitelist))

    def _del_all_keywords(self, table):
        """
        DB: Remove all keywords for current show

        :param table: SQL table remove keywords from
        """
        main_db_con = db.DBConnection()
        main_db_con.action("DELETE FROM [" + table + "] WHERE show_id = ?", [self.show_id])

    def _load_list(self, table):
        """
        DB: Fetch keywords for current show

        :param table: Table to fetch list of keywords from

        :return: keywords in list
        """
        main_db_con = db.DBConnection()
        sql_results = main_db_con.select("SELECT keyword FROM [" + table + "] WHERE show_id = ?", [self.show_id])
        if not sql_results or not len(sql_results):
            return []
        groups = []
        for result in sql_results:
            groups.append(result["keyword"])

        logger.debug("BWL: {id} loaded keywords from {table}: {groups}".format(id=self.show_id, table=table, groups=groups))

        return groups

    def is_valid(self, result):
        """
        Check if result is valid according to white/blacklist for current show

        :param result: Result to analyse
        :return: False if result is not allowed in white/blacklist, True if it is
        """

        if self.whitelist or self.blacklist:
            if not result.release_group:
                logger.debug("Failed to detect release group, invalid result")
                return False

            if result.release_group.lower() in [x.lower() for x in self.whitelist]:
                white_result = True
            elif not self.whitelist:
                white_result = True
            else:
                white_result = False
            if result.release_group.lower() in [x.lower() for x in self.blacklist]:
                black_result = False
            else:
                black_result = True

            logger.debug("Whitelist check passed: {white}. Blacklist check passed: {black}".format(white=white_result, black=black_result))

            if white_result and black_result:
                return True
            else:
                return False
        else:
            logger.debug("No Whitelist and Blacklist defined, check passed.")
            return True


class BlackWhitelistNoShowIDException(Exception):
    """No show_id was given"""


def short_group_names(groups):
    """
    Find AniDB short group names for release groups

    :param groups: list of groups to find short group names for
    :return: list of shortened group names
    """
    groups = groups.split(",")
    short_group_list = []
    if helpers.set_up_anidb_connection():
        for groupName in groups:
            try:
                group = settings.ADBA_CONNECTION.group(gname=groupName)
            except AniDBCommandTimeoutError:
                logger.debug("Timeout while loading group from AniDB. Trying next group")
            except Exception:
                logger.debug("Failed while loading group from AniDB. Trying next group")
            else:
                for line in group.datalines:
                    if line["shortname"]:
                        short_group_list.append(line["shortname"])
                    else:
                        if groupName not in short_group_list:
                            short_group_list.append(groupName)
    else:
        short_group_list = groups
    return short_group_list



import unittest

import sickchill.logger
from sickchill import settings
from sickchill.oldbeard.providers import bitcannon, rarbg
from sickchill.tv import TVEpisode, TVShow
from tests import conftest


class TorrentBasicTests(conftest.SickChillTestDBCase):
    """
    Test torrents
    """

    @classmethod
    def setUpClass(cls):
        cls.shows = []

        show = TVShow(1, 121361)
        show.name = "Italian Works"
        show.episodes = []
        episode = TVEpisode(show, 5, 10)
        episode.name = "Pines of Rome"
        episode.scene_season = 5
        episode.scene_episode = 10
        show.episodes.append(episode)
        cls.shows.append(show)

    def test_bitcannon(self):
        """
        Test bitcannon
        """
        provider = bitcannon.Provider()
        provider.custom_url = ""  # true testing requires a valid URL here (e.g., "http://localhost:3000/")
        provider.api_key = ""

        if provider.custom_url:

            search_strings_list = provider.get_episode_search_strings(self.shows[0].episodes[0])  # [{'Episode': ['Italian Works S05E10']}]
            for search_strings in search_strings_list:
                provider.search(search_strings)  # {'Episode': ['Italian Works S05E10']}

        return True

    @unittest.skip("Only run this manually")
    def test_search(self):
        """
        Test searching, using rarbg
        """
        settings.CPU_PRESET = "LOW"
        provider = rarbg.Provider()
        results = provider.search({"Episode": ["The Mandalorian S01E08"]})
        assert results
        self.assertIn("Mandalorian", results[0]["title"])

        results = provider.search({"RSS": [""]})
        assert results


if __name__ == "__main__":
    print("==================")
    print("STARTING - Torrent Basic TESTS")
    print("==================")
    print("######################################################################")

    def override_log(msg, *args, **kwargs):
        """Override the SickChill logger so we can see the debug output from providers"""
        _ = args, kwargs
        print(msg)

    sickchill.logger.info = override_log
    sickchill.logger.debug = override_log
    sickchill.logger.error = override_log
    sickchill.logger.warning = override_log

    SUITE = unittest.TestLoader().loadTestsFromTestCase(TorrentBasicTests)
    unittest.TextTestRunner(verbosity=2).run(SUITE)



# -*- coding: utf-8 -*-

# ------------------------------------------------------------------------------- #

"""
cloudscraper.exceptions
~~~~~~~~~~~~~~~~~~~
This module contains the set of cloudscraper exceptions.
"""

# ------------------------------------------------------------------------------- #


class CloudflareException(Exception):
    """
    Base exception class for cloudscraper for Cloudflare
    """


class CloudflareLoopProtection(CloudflareException):
    """
    Raise an exception for recursive depth protection
    """


class CloudflareCode1020(CloudflareException):
    """
    Raise an exception for Cloudflare code 1020 block
    """


class CloudflareIUAMError(CloudflareException):
    """
    Raise an error for problem extracting IUAM paramters
    from Cloudflare payload
    """


class CloudflareChallengeError(CloudflareException):
    """
    Raise an error when detected new Cloudflare challenge
    """


class CloudflareSolveError(CloudflareException):
    """
    Raise an error when issue with solving Cloudflare challenge
    """


class CloudflareCaptchaError(CloudflareException):
    """
    Raise an error for problem extracting Captcha paramters
    from Cloudflare payload
    """


class CloudflareCaptchaProvider(CloudflareException):
    """
    Raise an exception for no Captcha provider loaded for Cloudflare.
    """

# ------------------------------------------------------------------------------- #


class CaptchaException(Exception):
    """
    Base exception class for cloudscraper captcha Providers
    """


class CaptchaServiceUnavailable(CaptchaException):
    """
    Raise an exception for external services that cannot be reached
    """


class CaptchaAPIError(CaptchaException):
    """
    Raise an error for error from API response.
    """


class CaptchaAccountError(CaptchaException):
    """
    Raise an error for captcha provider account problem.
    """


class CaptchaTimeout(CaptchaException):
    """
    Raise an exception for captcha provider taking too long.
    """


class CaptchaParameter(CaptchaException):
    """
    Raise an exception for bad or missing Parameter.
    """


class CaptchaBadJobID(CaptchaException):
    """
    Raise an exception for invalid job id.
    """


class CaptchaReportError(CaptchaException):
    """
    Raise an error for captcha provider unable to report bad solve.
    """



import os

from tornado.web import addslash

import sickchill.start
from sickchill import logger, settings
from sickchill.helper import setup_github, try_int
from sickchill.init_helpers import setup_gettext
from sickchill.oldbeard import config, filters, helpers, ui
from sickchill.oldbeard.common import Quality, WANTED
from sickchill.views.common import PageTemplate
from sickchill.views.routes import Route

from .index import Config


@Route("/config/general(/?.*)", name="config:general")
class ConfigGeneral(Config):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @addslash
    def index(self):
        t = PageTemplate(rh=self, filename="config_general.mako")

        return t.render(
            title=_("Config - General"), header=_("General Configuration"), topmenu="config", submenu=self.ConfigMenu(), controller="config", action="index"
        )

    @staticmethod
    def generateApiKey():
        return helpers.generateApiKey()

    @staticmethod
    def saveRootDirs(rootDirString=None):
        settings.ROOT_DIRS = rootDirString

    @staticmethod
    def saveAddShowDefaults(
        defaultStatus, anyQualities, bestQualities, defaultSeasonFolders, subtitles=False, anime=False, scene=False, defaultStatusAfter=WANTED
    ):

        if anyQualities:
            anyQualities = anyQualities.split(",")
        else:
            anyQualities = []

        if bestQualities:
            bestQualities = bestQualities.split(",")
        else:
            bestQualities = []

        newQuality = Quality.combineQualities([int(quality) for quality in anyQualities], [int(quality) for quality in bestQualities])

        settings.STATUS_DEFAULT = int(defaultStatus)
        settings.STATUS_DEFAULT_AFTER = int(defaultStatusAfter)
        settings.QUALITY_DEFAULT = int(newQuality)

        settings.SEASON_FOLDERS_DEFAULT = config.checkbox_to_value(defaultSeasonFolders)
        settings.SUBTITLES_DEFAULT = config.checkbox_to_value(subtitles)

        settings.ANIME_DEFAULT = config.checkbox_to_value(anime)

        settings.SCENE_DEFAULT = config.checkbox_to_value(scene)
        sickchill.start.save_config()

        ui.notifications.message(_("Saved Defaults"), _('Your "add show" defaults have been set to your current selections.'))

    def saveGeneral(
        self,
        log_nr=5,
        log_size=1,
        web_port=None,
        notify_on_login=None,
        web_log=None,
        encryption_version=None,
        web_ipv6=None,
        trash_remove_show=None,
        trash_rotate_logs=None,
        update_frequency=None,
        skip_removed_files=None,
        indexerDefaultLang="en",
        ep_default_deleted_status=None,
        launch_browser=None,
        showupdate_hour=3,
        web_username=None,
        api_key=None,
        indexer_default=None,
        timezone_display=None,
        cpu_preset="NORMAL",
        web_password=None,
        version_notify=None,
        enable_https=None,
        https_cert=None,
        https_key=None,
        handle_reverse_proxy=None,
        sort_article=None,
        auto_update=None,
        notify_on_update=None,
        proxy_setting=None,
        proxy_indexers=None,
        anon_redirect=None,
        git_path=None,
        git_remote=None,
        calendar_unprotected=None,
        calendar_icons=None,
        debug=None,
        ssl_verify=None,
        no_restart=None,
        coming_eps_missed_range=None,
        fuzzy_dating=None,
        trim_zero=None,
        date_preset=None,
        date_preset_na=None,
        time_preset=None,
        indexer_timeout=None,
        download_url=None,
        rootDir=None,
        theme_name=None,
        default_page=None,
        fanart_background=None,
        fanart_background_opacity=None,
        sickchill_background=None,
        sickchill_background_path=None,
        custom_css=None,
        custom_css_path=None,
        git_reset=None,
        git_username=None,
        git_token=None,
        display_all_seasons=None,
        gui_language=None,
        ignore_broken_symlinks=None,
        ended_shows_update_interval=None,
    ):

        results = []

        if gui_language != settings.GUI_LANG:
            setup_gettext(gui_language)
            settings.GUI_LANG = gui_language

        # Misc
        settings.DOWNLOAD_URL = download_url
        settings.INDEXER_DEFAULT_LANGUAGE = indexerDefaultLang
        settings.EP_DEFAULT_DELETED_STATUS = ep_default_deleted_status
        settings.SKIP_REMOVED_FILES = config.checkbox_to_value(skip_removed_files)
        settings.LAUNCH_BROWSER = config.checkbox_to_value(launch_browser)
        config.change_showupdate_hour(showupdate_hour)
        config.change_version_notify(version_notify)
        settings.AUTO_UPDATE = config.checkbox_to_value(auto_update)
        settings.NOTIFY_ON_UPDATE = config.checkbox_to_value(notify_on_update)
        settings.LOG_NR = log_nr
        settings.LOG_SIZE = float(log_size)
        settings.WEB_LOG = config.checkbox_to_value(web_log)

        settings.TRASH_REMOVE_SHOW = config.checkbox_to_value(trash_remove_show)
        settings.TRASH_ROTATE_LOGS = config.checkbox_to_value(trash_rotate_logs)
        settings.IGNORE_BROKEN_SYMLINKS = config.checkbox_to_value(ignore_broken_symlinks)
        config.change_update_frequency(update_frequency)
        settings.LAUNCH_BROWSER = config.checkbox_to_value(launch_browser)
        settings.SORT_ARTICLE = config.checkbox_to_value(sort_article)
        settings.CPU_PRESET = cpu_preset
        settings.ANON_REDIRECT = anon_redirect
        settings.PROXY_SETTING = proxy_setting
        if settings.PROXY_SETTING:
            settings.PROXY_SETTING = config.clean_url(settings.PROXY_SETTING).rstrip("/")
        settings.PROXY_INDEXERS = config.checkbox_to_value(proxy_indexers)

        settings.GIT_USERNAME = git_username

        tmp_git_token = filters.unhide(settings.GIT_TOKEN, git_token)
        if settings.GIT_TOKEN != tmp_git_token:
            # Re-Initializes oldbeard.gh, so a restart isn't necessary
            settings.GIT_TOKEN = tmp_git_token
            setup_github()

        # oldbeard.GIT_RESET = config.checkbox_to_value(git_reset)
        # Force GIT_RESET
        settings.GIT_RESET = 1
        settings.GIT_PATH = git_path
        settings.GIT_REMOTE = git_remote
        settings.CALENDAR_UNPROTECTED = config.checkbox_to_value(calendar_unprotected)
        settings.CALENDAR_ICONS = config.checkbox_to_value(calendar_icons)
        settings.NO_RESTART = config.checkbox_to_value(no_restart)
        settings.DEBUG = config.checkbox_to_value(debug)
        logger.set_level()

        settings.SSL_VERIFY = config.checkbox_to_value(ssl_verify)
        helpers.set_opener(settings.SSL_VERIFY)

        settings.COMING_EPS_MISSED_RANGE = config.min_max(coming_eps_missed_range, 7, 0, 42810)

        settings.DISPLAY_ALL_SEASONS = config.checkbox_to_value(display_all_seasons)
        settings.NOTIFY_ON_LOGIN = config.checkbox_to_value(notify_on_login)
        settings.WEB_PORT = try_int(web_port)
        settings.WEB_IPV6 = config.checkbox_to_value(web_ipv6)
        settings.ENCRYPTION_VERSION = config.checkbox_to_value(encryption_version, value_on=2, value_off=0)
        settings.WEB_USERNAME = web_username
        settings.WEB_PASSWORD = filters.unhide(settings.WEB_PASSWORD, web_password)

        settings.FUZZY_DATING = config.checkbox_to_value(fuzzy_dating)
        settings.TRIM_ZERO = config.checkbox_to_value(trim_zero)

        if date_preset:
            settings.DATE_PRESET = date_preset

        if indexer_default:
            settings.INDEXER_DEFAULT = try_int(indexer_default)

        if indexer_timeout:
            settings.INDEXER_TIMEOUT = try_int(indexer_timeout)

        if time_preset:
            settings.TIME_PRESET_W_SECONDS = time_preset
            settings.TIME_PRESET = time_preset.replace(":%S", "")

        settings.TIMEZONE_DISPLAY = timezone_display

        settings.API_KEY = api_key

        settings.ENABLE_HTTPS = config.checkbox_to_value(enable_https)

        if not config.change_https_cert(https_cert):
            results += [_("Unable to create directory {directory}, https cert directory not changed.").format(directory=os.path.normpath(https_cert))]

        if not config.change_https_key(https_key):
            results += [_("Unable to create directory {directory}, https key directory not changed.").format(directory=os.path.normpath(https_key))]

        settings.HANDLE_REVERSE_PROXY = config.checkbox_to_value(handle_reverse_proxy)

        settings.THEME_NAME = theme_name
        settings.SICKCHILL_BACKGROUND = config.checkbox_to_value(sickchill_background)
        config.change_sickchill_background(sickchill_background_path)
        settings.FANART_BACKGROUND = config.checkbox_to_value(fanart_background)
        settings.FANART_BACKGROUND_OPACITY = fanart_background_opacity
        settings.CUSTOM_CSS = config.checkbox_to_value(custom_css)
        config.change_custom_css(custom_css_path)

        settings.ENDED_SHOWS_UPDATE_INTERVAL = int(ended_shows_update_interval)

        settings.DEFAULT_PAGE = default_page

        sickchill.start.save_config()

        if len(results) > 0:
            for x in results:
                logger.exception(x)
            ui.notifications.error(_("Error(s) Saving Configuration"), "<br>\n".join(results))
        else:
            ui.notifications.message(_("Configuration Saved"), os.path.join(settings.CONFIG_FILE))

        return self.redirect("/config/general/")



import sys

from . import generic, helpers, kodi, mede8er, mediabrowser, ps3, tivo, wdtv

__all__ = ["generic", "helpers", "kodi", "mede8er", "mediabrowser", "ps3", "tivo", "wdtv"]


def available_generators():
    return [x for x in __all__ if x not in ["generic", "helpers"]]


def _getMetadataModule(name):
    name = name.lower()
    prefix = "sickchill.providers.metadata."
    if name in available_generators() and prefix + name in sys.modules:
        return sys.modules[prefix + name]
    else:
        return None


def _getMetadataClass(name):
    module = _getMetadataModule(name)
    if not module:
        return None

    # noinspection PyUnresolvedReferences
    return module.metadata_class()


def get_metadata_generator_dict():
    result = {}
    for cur_generator_id in available_generators():
        cur_generator = _getMetadataClass(cur_generator_id)
        if not cur_generator:
            continue
        result[cur_generator.name] = cur_generator

    return result



import abc
import logging
import sys

if sys.version_info >= (3, 4):
    ABC = abc.ABC  # noqa
else:
    ABC = abc.ABCMeta('ABC', (), {})

# ------------------------------------------------------------------------------- #

captchaSolvers = {}

# ------------------------------------------------------------------------------- #


class Captcha(ABC):
    @abc.abstractmethod
    def __init__(self, name):
        captchaSolvers[name] = self

    # ------------------------------------------------------------------------------- #

    @classmethod
    def dynamicImport(cls, name):
        if name not in captchaSolvers:
            try:
                __import__(f'{cls.__module__}.{name}')
                if not isinstance(captchaSolvers.get(name), Captcha):
                    raise ImportError('The anti captcha provider was not initialized.')
            except ImportError as e:
                sys.tracebacklimit = 0
                logging.error(f'Unable to load {name} anti captcha provider -> {e}')
                raise

        return captchaSolvers[name]

    # ------------------------------------------------------------------------------- #

    @abc.abstractmethod
    def getCaptchaAnswer(self, captchaType, url, siteKey, captchaParams):
        pass

    # ------------------------------------------------------------------------------- #

    def solveCaptcha(self, captchaType, url, siteKey, captchaParams):
        return self.getCaptchaAnswer(captchaType, url, siteKey, captchaParams)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2007-2018 The Python-Twitter Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A library that provides a Python interface to the Twitter API."""
from __future__ import absolute_import

__author__       = 'The Python-Twitter Developers'
__email__        = 'python-twitter@googlegroups.com'
__copyright__    = 'Copyright (c) 2007-2016 The Python-Twitter Developers'
__license__      = 'Apache License 2.0'
__version__      = '3.5'
__url__          = 'https://github.com/bear/python-twitter'
__download_url__ = 'https://pypi.python.org/pypi/python-twitter'
__description__  = 'A Python wrapper around the Twitter API'


import json                                 # noqa

try:
    from hashlib import md5                 # noqa
except ImportError:
    from md5 import md5                     # noqa

from ._file_cache import _FileCache         # noqa
from .error import TwitterError             # noqa
from .parse_tweet import ParseTweet         # noqa

from .models import (                       # noqa
    Category,                               # noqa
    DirectMessage,                          # noqa
    Hashtag,                                # noqa
    List,                                   # noqa
    Media,                                  # noqa
    Trend,                                  # noqa
    Url,                                    # noqa
    User,                                   # noqa
    UserStatus,                             # noqa
    Status                                  # noqa
)

from .api import Api                        # noqa



import json
import os
import random
import re
import sys
import ssl

from collections import OrderedDict

# ------------------------------------------------------------------------------- #


class User_Agent():

    # ------------------------------------------------------------------------------- #

    def __init__(self, *args, **kwargs):
        self.headers = None
        self.cipherSuite = []
        self.loadUserAgent(*args, **kwargs)

    # ------------------------------------------------------------------------------- #

    def filterAgents(self, user_agents):
        filtered = {}

        if self.mobile:
            if self.platform in user_agents['mobile'] and user_agents['mobile'][self.platform]:
                filtered.update(user_agents['mobile'][self.platform])

        if self.desktop:
            if self.platform in user_agents['desktop'] and user_agents['desktop'][self.platform]:
                filtered.update(user_agents['desktop'][self.platform])

        return filtered

    # ------------------------------------------------------------------------------- #

    def tryMatchCustom(self, user_agents):
        for device_type in user_agents['user_agents']:
            for platform in user_agents['user_agents'][device_type]:
                for browser in user_agents['user_agents'][device_type][platform]:
                    if re.search(re.escape(self.custom), ' '.join(user_agents['user_agents'][device_type][platform][browser])):
                        self.headers = user_agents['headers'][browser]
                        self.headers['User-Agent'] = self.custom
                        self.cipherSuite = user_agents['cipherSuite'][browser]
                        return True
        return False

    # ------------------------------------------------------------------------------- #

    def loadUserAgent(self, *args, **kwargs):
        self.browser = kwargs.pop('browser', None)

        self.platforms = ['linux', 'windows', 'darwin', 'android', 'ios']
        self.browsers = ['chrome', 'firefox']

        if isinstance(self.browser, dict):
            self.custom = self.browser.get('custom', None)
            self.platform = self.browser.get('platform', None)
            self.desktop = self.browser.get('desktop', True)
            self.mobile = self.browser.get('mobile', True)
            self.browser = self.browser.get('browser', None)
        else:
            self.custom = kwargs.pop('custom', None)
            self.platform = kwargs.pop('platform', None)
            self.desktop = kwargs.pop('desktop', True)
            self.mobile = kwargs.pop('mobile', True)

        if not self.desktop and not self.mobile:
            sys.tracebacklimit = 0
            raise RuntimeError("Sorry you can't have mobile and desktop disabled at the same time.")

        with open(os.path.join(os.path.dirname(__file__), 'browsers.json'), 'r') as fp:
            user_agents = json.load(
                fp,
                object_pairs_hook=OrderedDict
            )

        if self.custom:
            if not self.tryMatchCustom(user_agents):
                self.cipherSuite = [
                    ssl._DEFAULT_CIPHERS,
                    '!AES128-SHA',
                    '!ECDHE-RSA-AES256-SHA',
                ]
                self.headers = OrderedDict([
                    ('User-Agent', self.custom),
                    ('Accept', 'text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8'),
                    ('Accept-Language', 'en-US,en;q=0.9'),
                    ('Accept-Encoding', 'gzip, deflate, br')
                ])
        else:
            if self.browser and self.browser not in self.browsers:
                sys.tracebacklimit = 0
                raise RuntimeError(f'Sorry "{self.browser}" browser is not valid, valid browsers are [{", ".join(self.browsers)}].')

            if not self.platform:
                self.platform = random.SystemRandom().choice(self.platforms)

            if self.platform not in self.platforms:
                sys.tracebacklimit = 0
                raise RuntimeError(f'Sorry the platform "{self.platform}" is not valid, valid platforms are [{", ".join(self.platforms)}]')

            filteredAgents = self.filterAgents(user_agents['user_agents'])

            if not self.browser:
                # has to be at least one in there...
                while not filteredAgents.get(self.browser):
                    self.browser = random.SystemRandom().choice(list(filteredAgents.keys()))

            if not filteredAgents[self.browser]:
                sys.tracebacklimit = 0
                raise RuntimeError(f'Sorry "{self.browser}" browser was not found with a platform of "{self.platform}".')

            self.cipherSuite = user_agents['cipherSuite'][self.browser]
            self.headers = user_agents['headers'][self.browser]

            self.headers['User-Agent'] = random.SystemRandom().choice(filteredAgents[self.browser])

        if not kwargs.get('allow_brotli', False) and 'br' in self.headers['Accept-Encoding']:
            self.headers['Accept-Encoding'] = ','.join([
                encoding for encoding in self.headers['Accept-Encoding'].split(',') if encoding.strip() != 'br'
            ]).strip()



from fanart.items import Immutable, LeafItem, ResourceItem, CollectableItem
import fanart
__all__ = (
    'BackgroundItem',
    'CoverItem',
    'LogoItem',
    'ThumbItem',
    'DiscItem',
    'Artist',
    'Album',
)


class BackgroundItem(LeafItem):
    KEY = fanart.TYPE.MUSIC.BACKGROUND


class CoverItem(LeafItem):
    KEY = fanart.TYPE.MUSIC.COVER


class LogoItem(LeafItem):
    KEY = fanart.TYPE.MUSIC.LOGO


class ThumbItem(LeafItem):
    KEY = fanart.TYPE.MUSIC.THUMB


class DiscItem(LeafItem):
    KEY = fanart.TYPE.MUSIC.DISC

    @Immutable.mutablemethod
    def __init__(self, id, url, likes, disc, size):
        super(DiscItem, self).__init__(id, url, likes)
        self.disc = int(disc)
        self.size = int(size)


class Artist(ResourceItem):
    WS = fanart.WS.MUSIC

    @Immutable.mutablemethod
    def __init__(self, name, mbid, albums, backgrounds, logos, thumbs):
        self.name = name
        self.mbid = mbid
        self.albums = albums
        self.backgrounds = backgrounds
        self.logos = logos
        self.thumbs = thumbs

    @classmethod
    def from_dict(cls, resource):
        minimal_keys = {'name', 'mbid_id'}
        assert all(k in resource for k in minimal_keys), 'Bad Format Map'
        return cls(
            name=resource['name'],
            mbid=resource['mbid_id'],
            albums=Album.collection_from_dict(resource.get('albums', {})),
            backgrounds=BackgroundItem.extract(resource),
            thumbs=ThumbItem.extract(resource),
            logos=LogoItem.extract(resource),
        )


class Album(CollectableItem):

    @Immutable.mutablemethod
    def __init__(self, mbid, covers, arts):
        self.mbid = mbid
        self.covers = covers
        self.arts = arts

    @classmethod
    def from_dict(cls, key, resource):
        return cls(
            mbid=key,
            covers=CoverItem.extract(resource),
            arts=DiscItem.extract(resource),
        )



#
##############################################################################
import ast
import socket
import time
from http.client import HTTPException, HTTPSConnection
from ssl import SSLError
from urllib.parse import urlencode

from sickchill import logger, settings
from sickchill.oldbeard import common, db


class Notifier(object):
    def test_notify(self, prowl_api, prowl_priority):
        return self._send_prowl(prowl_api, prowl_priority, event="Test", message="Testing Prowl settings from SickChill", force=True)

    def notify_snatch(self, ep_name):
        if settings.PROWL_NOTIFY_ONSNATCH:
            show = self._parse_episode(ep_name)
            recipients = self._generate_recipients(show)
            if not recipients:
                logger.debug("Skipping prowl notify because there are no configured recipients")
            else:
                for api in recipients:
                    self._send_prowl(
                        prowl_api=api,
                        prowl_priority=None,
                        event=common.notifyStrings[common.NOTIFY_SNATCH],
                        message=ep_name + " :: " + time.strftime(settings.DATE_PRESET + " " + settings.TIME_PRESET),
                    )

    def notify_download(self, ep_name):
        if settings.PROWL_NOTIFY_ONDOWNLOAD:
            show = self._parse_episode(ep_name)
            recipients = self._generate_recipients(show)
            if not recipients:
                logger.debug("Skipping prowl notify because there are no configured recipients")
            else:
                for api in recipients:
                    self._send_prowl(
                        prowl_api=api,
                        prowl_priority=None,
                        event=common.notifyStrings[common.NOTIFY_DOWNLOAD],
                        message=ep_name + " :: " + time.strftime(settings.DATE_PRESET + " " + settings.TIME_PRESET),
                    )

    def notify_subtitle_download(self, ep_name, lang):
        if settings.PROWL_NOTIFY_ONSUBTITLEDOWNLOAD:
            show = self._parse_episode(ep_name)
            recipients = self._generate_recipients(show)
            if not recipients:
                logger.debug("Skipping prowl notify because there are no configured recipients")
            else:
                for api in recipients:
                    self._send_prowl(
                        prowl_api=api,
                        prowl_priority=None,
                        event=common.notifyStrings[common.NOTIFY_SUBTITLE_DOWNLOAD],
                        message=ep_name + " [" + lang + "] :: " + time.strftime(settings.DATE_PRESET + " " + settings.TIME_PRESET),
                    )

    def notify_git_update(self, new_version="??"):
        if settings.USE_PROWL:
            update_text = common.notifyStrings[common.NOTIFY_GIT_UPDATE_TEXT]
            title = common.notifyStrings[common.NOTIFY_GIT_UPDATE]
            self._send_prowl(prowl_api=None, prowl_priority=None, event=title, message=update_text + new_version)

    def notify_login(self, ipaddress=""):
        if settings.USE_PROWL:
            update_text = common.notifyStrings[common.NOTIFY_LOGIN_TEXT]
            title = common.notifyStrings[common.NOTIFY_LOGIN]
            self._send_prowl(prowl_api=None, prowl_priority=None, event=title, message=update_text.format(ipaddress))

    @staticmethod
    def _generate_recipients(show=None):
        apis = []
        mydb = db.DBConnection(row_type="dict")

        # Grab the global recipient(s)
        if settings.PROWL_API:
            for api in settings.PROWL_API.split(","):
                if api.strip():
                    apis.append(api.strip())

        # Grab the per-show-notification recipients
        if show is not None:
            for value in show:
                for subs in mydb.select("SELECT notify_list FROM tv_shows WHERE show_name = ?", (value,)):
                    if subs["notify_list"] and subs["notify_list"][0] == "{":  # legacy format handling
                        entries = dict(ast.literal_eval(subs["notify_list"]))
                        for api in entries["prowlAPIs"].split(","):
                            if api.strip():
                                apis.append(api.strip())

        apis = set(apis)
        return apis

    @staticmethod
    def _send_prowl(prowl_api=None, prowl_priority=None, event=None, message=None, force=False):

        if not settings.USE_PROWL and not force:
            return False

        if prowl_api is None:
            prowl_api = settings.PROWL_API
            if len(prowl_api) == 0:
                return False

        if prowl_priority is None:
            prowl_priority = settings.PROWL_PRIORITY

        title = settings.PROWL_MESSAGE_TITLE

        logger.debug(
            'PROWL: Sending notice with details: title="{0}" event="{1}", message="{2}", priority={3}, api={4}'.format(
                title, event, message, prowl_priority, prowl_api
            )
        )

        http_handler = HTTPSConnection("api.prowlapp.com")

        data = {"apikey": prowl_api, "application": title, "event": event, "description": message, "priority": prowl_priority}

        try:
            http_handler.request("POST", "/publicapi/add", headers={"Content-type": "application/x-www-form-urlencoded"}, body=urlencode(data))
        except (SSLError, HTTPException, socket.error):
            logger.exception("Prowl notification failed.")
            return False
        response = http_handler.getresponse()
        request_status = response.status

        if request_status == 200:
            logger.info("Prowl notifications sent.")
            return True
        elif request_status == 401:
            logger.exception("Prowl auth failed: {0}".format(response.reason))
            return False
        else:
            logger.exception("Prowl notification failed.")
            return False

    @staticmethod
    def _parse_episode(ep_name):
        sep = " - "
        titles = ep_name.split(sep)
        titles.sort(key=len, reverse=True)
        logger.debug("TITLES: {0}".format(titles))
        return titles



import logging

from sickchill import settings
from sickchill.oldbeard import config

from .common import PageTemplate
from .index import WebRoot

logger = logging.getLogger("sickchill.movie")


class MoviesHandler(WebRoot):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _genericMessage(self, subject=None, message=None):
        t = PageTemplate(rh=self, filename="genericMessage.mako")
        return t.render(message=message, subject=subject, topmenu="movies", title="")

    def index(self):
        t = PageTemplate(rh=self, filename="movies/index.mako")
        return t.render(title=_("Movies"), header=_("Movie List"), topmenu="movies", movies=settings.movie_list, controller="movies", action="index")

    def search(self):
        query = self.get_body_argument("query", "")
        year = self.get_body_argument("year", "")
        language = self.get_body_argument("language", "")
        adult = config.checkbox_to_value(self.get_body_argument("adult", False))
        search_results = []
        if query:
            search_results = settings.movie_list.search_tmdb(query=query, year=year, language=language, adult=adult)
        t = PageTemplate(rh=self, filename="movies/search.mako")
        return t.render(
            title=_("Movies"),
            header=_("Movie Search"),
            topmenu="movies",
            controller="movies",
            action="search",
            search_results=search_results,
            movies=settings.movie_list,
            query=query,
            year=year,
            language=language,
            adult=adult,
        )

    def add(self):
        movie = None
        imdb_id = self.get_body_argument("imdb", None)
        if imdb_id:
            movie = settings.movie_list.add_from_imdb(imdb_id=imdb_id)
        tmdb_id = self.get_body_argument("tmdb", None)
        if tmdb_id:
            movie = settings.movie_list.add_from_tmdb(tmdb_id=tmdb_id)

        if not movie:
            return self.redirect(self.reverse_url("movies-search", "search"))

        return self.redirect(self.reverse_url("movies-details", "details", movie.slug))

    def remove(self):
        pk = self.path_kwargs.get("pk")
        if pk is not None:
            if not settings.movie_list.query.get(pk):
                return self._genericMessage(_("Error"), _("Movie not found"))

            settings.movie_list.delete(pk)

        t = PageTemplate(rh=self, filename="movies/remove.mako")
        return t.render(title=_("Movies"), header=_("Movie Remove"), topmenu="movies", movies=settings.movie_list, controller="movies", action="remove")

    def details(self):
        movie = settings.movie_list.by_slug(self.path_kwargs.get("slug"))
        if not movie:
            return self._genericMessage(_("Error"), _("Movie not found"))

        t = PageTemplate(rh=self, filename="movies/details.mako")
        return t.render(title=_("Movies"), header=_("Movie Remove"), topmenu="movies", controller="movies", action="details", movie=movie, movie_message=None)



# -*- coding: utf-8 -*-

"""
This module implements the User functionality of TheTVDb API.
Allows to retrieve, add and delete user favorites and ratings.

See [Users API section](https://api.thetvdb.com/swagger#!/Users)
"""

from .base import TVDB

class User(TVDB):
    """
    User class to retrieve, add and delete user favorites and ratings.
    Requires username and user-key.
    """
    _BASE_PATH = 'user'
    _URLS = {
        'info': '',
        'favorites': '/favorites',
        'alter_favorite': '/favorites/{id}'
    }

    def __init__(self, user, key):
        """
        Initialize the User class.

        `user` is the username for login. `key` is the userkey needed to 
        authenticate with the user, you can find it in the 
        [account info](http://thetvdb.com/?tab=userinfo) under account identifier.
        """
        super(User, self).__init__(user=user, key=key)
        self.Ratings = User_Ratings(user, key)
        """
        Allows to retrieve, add and delete user ratings.
        """

    def info(self):
        """
        Get the basic user info and set its values to local attributes.
        
        Returns a dict with all the information of the user.

        For example

            #!python
            >>> import tvdbsimple as tvdb
            >>> tvdb.KEYS.API_KEY = 'YOUR_API_KEY'
            >>> user = tvdb.User('username', 'userkey')
            >>> response = user.info()
            >>> user.userName
            'username'

        """
        path = self._get_path('info')
        
        response = self._GET(path)
        self._set_attrs_to_values(response)
        return response

    def favorites(self):
        """
        Get the a list of the favorite series of the user and
        sets it to `favorites` attribute.
        
        Returns a list of the favorite series ids.

        For example

            #!python
            >>> import tvdbsimple as tvdb
            >>> tvdb.KEYS.API_KEY = 'YOUR_API_KEY'
            >>> user = tvdb.User('username', 'userkey')
            >>> response = user.favorites()
            >>> user.favorites[0]
            '73545'

        """
        path = self._get_path('favorites')
        
        response = self._GET(path)
        self._set_attrs_to_values(response)
        return self._clean_return(response)

    def _clean_return(self,jsn):
        if 'favorites' in jsn:
            return jsn['favorites']
        return jsn

    def add_favorite(self, id):
        """
        Add a series to user favorite series from its series id.
        
        `id` is the series id you want to add to favorites.

        Returns the updated list of the favorite series ids.

        For example

            #!python
            >>> import tvdbsimple as tvdb
            >>> tvdb.KEYS.API_KEY = 'YOUR_API_KEY'
            >>> user = tvdb.User('username', 'userkey')
            >>> response = user.add_favorite(78804)
            >>> response[-1]
            '78804'

        """
        path = self._get_path('alter_favorite').format(id=id)
        
        return self._clean_return(self._PUT(path))

    def delete_favorite(self, id):
        """
        Delete a series from user favorite series from its series id.
        
        `id` is the series id you want to delete from favorites.

        Returns the updated list of the favorite series ids.

        For example

            #!python
            >>> import tvdbsimple as tvdb
            >>> tvdb.KEYS.API_KEY = 'YOUR_API_KEY'
            >>> user = tvdb.User('username', 'userkey')
            >>> response = user.delete_favorite(78804)
            >>> response[-1]
            '73545'

        """
        path = self._get_path('alter_favorite').format(id=id)
        
        return self._clean_return(self._DELETE(path))

class User_Ratings(TVDB):
    """
    Class needed to organize user ratings. Allows to retrieve, add and delete user ratings.

    Requires username and user-key.
    """
    _BASE_PATH = 'user/ratings'
    _URLS = {
        'all': '',
        'query': '/query',
        'query_params': '/query/params',
        'add': '/{itemType}/{itemId}/{itemRating}',
        'delete': '/{itemType}/{itemId}'
    }

    def __init__(self, user, key, **kwargs):
        """
        Initialize the class.

        `user` is the username for login. `key` is the userkey needed to 
        authenticate with the user, you can find it in the 
        [account info](http://thetvdb.com/?tab=userinfo) under account identifier.
        It's possible to provide `itemType` that filters ratings by type. 
        Can be either 'series', 'episode', or 'banner'.
        """
        super(User_Ratings, self).__init__(user=user, key=key)
        self.update_filters(**kwargs)

    def update_filters(self, **kwargs):
        """
        Set the filters for the user rating.

        It's possible to provide `itemType` that filters ratings by type. 
        Can be either 'series', 'episode', or 'banner'.
        """
        self._PAGES = -1
        self._PAGES_LIST = {}
        self._FILTERS = kwargs

    def query_params(self):
        """
        Get the query parameters allowed for filtering and set it to `query_params` attribute.

        Returns a list of parameters you can set to filters.
        """
        path = self._get_id_path('query_params')
        
        response = self._GET(path)
        self._set_attrs_to_values({'query_params': response})
        return response

    def pages(self):
        """
        Get the number of rating pages available for filtered ratings of the specific user.

        Returns the number of rating pages available with current filters.
        """
        if self._PAGES < 0:
            self.page(1)
        return self._PAGES

    def add(self, type, id, rating):
        """
        Add a new rating to the user's ratings..
        
        `type` is the item type of the item you want to rate. Can be either 
        'series', 'episode', or 'image'.
        `id` is the ID of the item that you want to rate.
        `rating` is the `integer` rating you want to set.

        Returns a list with the new updated rating.

        For example

            #!python
            >>> import tvdbsimple as tvdb
            >>> tvdb.KEYS.API_KEY = 'YOUR_API_KEY'
            >>> rtn = tvdb.User_Ratings('username', 'userkey')
            >>> response = rtn.add('series', 78804, 8)

        """
        path = self._get_path('add').format(itemType=type, itemId=id, itemRating=rating)
        
        return self._PUT(path)

    def delete(self, type, id):
        """
        Delete an existing user's rating..
        
        `type` is the item type of the item rating you want to delete. Can be either 
        'series', 'episode', or 'image'.
        `id` is the ID of the item rating that you want to delete.

        Returns an empty dictionary.

        For example

            #!python
            >>> import tvdbsimple as tvdb
            >>> tvdb.KEYS.API_KEY = 'YOUR_API_KEY'
            >>> rtn = tvdb.User_Ratings('username', 'userkey')
            >>> response = rtn.delete('series', 78804)

        """
        path = self._get_path('delete').format(itemType=type, itemId=id)
        
        return self._DELETE(path)

    def all(self):
        """
        Get the full rating list filtered for the user and adds it
        to the `ratings` attribute.
        
        Returns a list of ratings info.

        For example

            #!python
            >>> import tvdbsimple as tvdb
            >>> tvdb.KEYS.API_KEY = 'YOUR_API_KEY'
            >>> rtn = tvdb.User_Rating('phate89', '3EF7CF9BBC8BB430')
            >>> response = rtn.all()
            >>> rtn.ratings[0]['ratingType']
            'episode'

        """
        ratings = []
        for i in range (1, self.pages()+1):
            ratings.extend(self.page(i))
        
        self._set_attrs_to_values({'ratings': ratings})
        return ratings

    def page(self, page):
        """
        Get the rating list for a specific page for the user.
        
        `page` is the rating page number.

        Returns a list ratings available in the page.
        """
        if page in self._PAGES_LIST:
            return self._PAGES_LIST[page]
        if self._FILTERS:
            path = self._get_path('query')
        else:
            path = self._get_path('all')
        
        filters = self._FILTERS.copy()
        filters['page'] = page
        response = self._GET(path, params=filters, cleanJson=False)
        if 'links' in response and 'last' in response['links']:
            self._PAGES = response['links']['last']

        self._PAGES_LIST[page] = response['data']
        return response['data']

    def __iter__(self):
        for i in range (1, self.pages()+1):
            yield self.page(i)


import os
import unittest
from httpretty import HTTPretty, httprettified
from fanart.music import *
from fanart.tests import LOCALDIR
os.environ['FANART_APIKEY'] = 'e3c7f0d0beeaf45b3a0dd3b9dd8a3338'


class ArtistItemTestCase(unittest.TestCase):
    @httprettified
    def test_get(self):
        artist_id = '24e1b53c-3085-4581-8472-0b0088d2508c'
        with open(os.path.join(LOCALDIR, 'response/music_a7f.json')) as fp:
            body = fp.read()
        HTTPretty.register_uri(
            HTTPretty.GET,
            'http://webservice.fanart.tv/v3/music/{}?api_key={}'.format(
                artist_id, os.environ['FANART_APIKEY']),
            body=body
        )
        a7f = Artist.get(id=artist_id)
        self.assertEqual(a7f.mbid, artist_id)
        self.assertEqual(a7f, eval(repr(a7f)))
        self.assertEqual(len(a7f.thumbs), 4)
